#define MDIM 2
#define SKALAR double

#define MAXTEIL	8
#define MAXPKT	100   /* ursprgl.: 500 */

typedef struct
{
   char pnam[4];	/* Parameternamen */
   char *expr;		/* Verweis auf String mit Funktionsausdruck */
} Mfkt;

typedef struct	  /* Definitionsgroessen fuer Multiquadrik (nach M. Eck) */
{
   SKALAR   rad;	/* globaler Faktor */
   SKALAR   zmin[MAXTEIL], zmax[MAXTEIL],
	    scal[MAXTEIL],
	    p[MAXTEIL][MAXPKT][3],
	    a[MAXTEIL][MAXPKT],
	    r[MAXTEIL][MAXPKT];
   int	    ianz[MAXTEIL];
   short    tlg;	/* Teilung der Koeffizientenmatrizen */
} TabMQ;

typedef struct
{
   SKALAR   *km,	      /* Koeffizientenmatrix */
	    *b0, *b1, *b2;    /* Hilfsfelder fuer Casteljau-Algorithmus */
   short    mg[3];	      /* maximale Polynomgrade */
} TabFP;

typedef struct Value
{
   short    type;	/* 1: skalar, 2: komplex bei Werten;
			   #Parameter bei integrierten Funktionen
			   MB, BB, MQ bei tabellierten Funktionen */
   union
   {
      SKALAR	s;	    /* skalare Werte */
      SKALAR	v[MDIM];    /* nicht-skalare Werte */
      char	lv;	    /* logischer Wert */
      Mfkt      mf;	    /* Makro oder Funktion */
      void	(*f)();	    /* Funktionen    */
      TabFP	*tfp;	    /* tabellierte Funktion, polynomial */
      TabMQ	*tmq;	    /* tabellierte Funktion, Multiquadrik */
   } u;
} Value;

/* Datenstruktur von Value:
 * Typ	 Art		Wert
 * ---   ---		----
 * -1	 logischer Wert	lv
 *  1	 reelle Zahl	s
 *  2	 komplexe Zahl	Re, Im
 */

typedef struct Symbol	/* Symboltabelle */
{
   char	    *name;
   struct Symbol *next;	/* Verkettung */
   short    class;	/* CONST, VAR, UNDEF, BLTIN, MFKT, TABF */
   Value    V;		/* Wert       */
} Symbol;

typedef struct Psymbol	/* Parameter-Symboltabelle */
{
   char	    pnam;
   Value    pval;		/* Wert       */
   struct Psymbol *next;	/* Verkettung */
} Psymbol;

Symbol *install();
char *emalloc();

typedef struct
{
   char *name;
   SKALAR cval;
}  Const_1;

#define lookup(s)	(Symbol *) tlookup(S_tab,s)
#define input() F_input()
#define unput(c) F_unput(c)
#define execerror(s,t) warning(s,t)

#define MULTPAR -1
#define MB	-2
#define BB	-3
#define MQ	-4
#define MAXFAKULT 33	/* groesstes Argument fuer Fakultaet */
#define EPSD 1e-10
#define EOS  '\0'

#ifndef NULL
#define NULL (char *)0
#endif

#ifdef vms
#define Pow Pow_
#define Cabs Cabs_
#define Abs Abs_
#define Ln Ln_
#define Exp Exp_
#define Sqrt Sqrt_
#define Sin Sin_
#define Cos Cos_
#define Tan Tan_
#define Sinh Sinh_
#define Cosh Cosh_
#define Tanh Tanh_
#define Asin Asin_
#define Acos Acos_
#define Atan Atan_
#define Atan2 Atan2_
#define Hypot Hypot_
#endif
