/*
 *				t a b l e s . c
 */

/*)LIBRARY
*/

/*
 *)EDITLEVEL=30
 * Edit history
 * 0.0 21-May-81 JSL	Invention
 * 0.1 22-May-81 JSL	Use signed arithmetic for speed; bug fixed; add tnext.
 * 1.0 28-May-81 JSL	Bugs arising from oddball unsigned conversions finally
 *			fixed; clean up user interface; add documentation.
 * 1.1 23-Jun-81 JSL	Improve (and fix) documentation; declare malloc().
 * 1.2 13-Jul-82 JSL	Removed extra argument from tdelete().
 */

#define NULL 0

#define TOOBIG 8192		/* Table size limit */
/* Define TOOBIG so that TOOBIG * sizeof(PAIR) just overflows an int */

#define EMPTY 0			/* Empty table entry key value */
#define DELETED 1		/* Deleted table entry value */
/*
 * EMPTY and DELETED must be two different values each distinct from any
 * possible string address.
 *
 */

#define LIMIT 5			/* "Random probe" limit */

#define streq(a,b)  strcmp(a,b)==0

extern char *malloc();
extern char *savestr();

typedef struct			/* One table entry */
{	char *key;
	int value;
} PAIR;

typedef struct
{	int tsize;
	PAIR tdata[100];
} TABLE;

TABLE *				/* Create a new table */
table(size)
register int size;
{	register TABLE *t;
	register int i;

	if ( size < 0 || size >= TOOBIG
	  || (t = (TABLE *) malloc(size*sizeof(PAIR)+sizeof(int))) == NULL )
		return(NULL);

	t->tsize = size;
	for (i = 0; i < size; i++)
		t->tdata[i].key = EMPTY;
	return(t);
}

PAIR *
__locate(t,key)
TABLE *t;
char *key;
{	register int bigh, i;
	register char *p;
	int h, size;
	char *free;
	unsigned long temp;

	bigh = 0;
	for (p = key; *p; )
		bigh = 253*bigh + *p++;
	free = NULL;
	size = t->tsize;

	for (i = 0; i < LIMIT; i++)		/* Random probes */
	{	bigh = 251 * bigh + 1233;
		temp = (unsigned)bigh;		/* Avoid compiler bug... */
		h = (temp * size) >> 16;
                h = h % size;
#ifdef DEBUG
		printf("Probe %d: bigh = %d[%u], h = %d[%u], size = %d, \
			temp = %ld\n",i,bigh,bigh,h,h,size,temp);
#endif
		p = t->tdata[h].key;
		if (free == NULL && (p == EMPTY || p == DELETED))
			free = &t->tdata[h];
		if (p == EMPTY)
			return(free);
		if (p != DELETED && streq(p,key))
			return(&t->tdata[h]);
	}

#ifdef DEBUG
	printf("Random probes failed...\n");
#endif

	for (h = 0; h < size; h++)		/* Exhaustive search */
	{	p = t->tdata[h].key;
		if (free == NULL && (p == EMPTY || p == DELETED))
			free = &t->tdata[h];
		if (p == EMPTY)
			return(free);
		if (p != DELETED && streq(p,key))
			return(&t->tdata[h]);
	}
	return(free);
}

tdelete(t,key)
TABLE *t;
char *key;
{	register PAIR *p;

	p = __locate(t,key);
	if (p == NULL || p->key == EMPTY || p->key == DELETED)
		return(1);
	free(p->key);
	p->key = DELETED;
	return(0);
}

tinsert(t,key,value)
TABLE *t;
char *key;
int value;
{	register PAIR *p;

	p = __locate(t,key);
	if (p == NULL)					/* Table overflow */
		return(-1);
	if (p->key != EMPTY && p->key != DELETED)	/* It's there */
		return(1);
	if ((p->key = savestr(key)) == NULL)		/* No core */
		return(-1);
	p->value = value;
	return(0);
}

tlookup(t,key)
TABLE *t;
char *key;
{	register PAIR *p;

	p = __locate(t,key);
	if (p == NULL || p->key == EMPTY || p->key == DELETED)
		return(0);
	return(p->value);
}

tnext(t,n,key,value)
TABLE *t;
int n;
char **key;
int *value;
{	register int i, size;
	register PAIR *p;

	if ( n < 0 )
		n = -1;

	size = t->tsize;
	for (i = n + 1; i < size; i++)
		if ((p = t->tdata[i].key) != EMPTY && p != DELETED)
		{	*key = p;
			*value = t->tdata[i].value;
			return(i);
		}
	return(-1);
}

tupdate(t,key,value)
TABLE *t;
char *key;
int value;
{	register PAIR *p;

	p = __locate(t,key);
	if (p == NULL || p->key == EMPTY || p->key == DELETED)
		return(1);
	p->value = value;
	return(0);
}
