#include <stdio.h>
#include "tm.h"
#include "nprime.c"

#define	 MAXLK 7	      /* maximale Kettenlaenge */

TMelem *TMA_ele(),
       *TMI_ele();

TMtab *			      /* Verweis auf erzeugte Tabelle */
TMS_def (tdim, smax)	      /* Erzeugen einer Koordinaten-Tabelle */
short tdim;		      /* (E) Dimension der Koordinaten */
TMkey smax[];		      /* (E) max. Aufloesung pro Dimension */
{
    int i, 
	size;
    short dim;
    TMtab *t = NULL;

    dim = ( tdim > KDIM ? KDIM : tdim );

    if (dim > 0)
    {
	for (size = 1, i = 0;  i < dim;  i++)
	    size *= smax[i];

	size = next_prime(size);

	if (size > 0 && size <= MAXNELEM &&
	    (t = malloc(2 * sizeof(int) + 3 * sizeof(unsigned short) +
	         KDIM * sizeof(TMkey) + sizeof(TMelist *))))
	{
	    t->tsize = size;
	    t->tdim  = dim;
	    t->typ   = tdim;
	    t->maxLk = MAXLK;
	    t->wref  = 0;

	    for (i = 0;  i < dim;  i++)
		t->tmax[i] = smax[i];

	    if (t->tm_elist = malloc(size * sizeof(TMelem *)))
	       memset(t->tm_elist->tm_elem, 0, size * sizeof(TMelem *));
/* oder (langsam!):
	       for (i = 0;  i < size;  i++)
		  t->tm_elist->tm_elem[i] = LEER;
*/
	    else
	    {
	       free(t);
	       t = NULL;
	    }
	}
    }

    return (t);
}

TMelem *		/* Verweis auf eingefuegtes Element (falls ir >= 0) */
TMS_ins (t, key, ir)	/* Einfuegen eines Elements in die Tabelle */
TMtab	*t;		/* (E) Verweis auf aktuelle Tabelle */
TMkey	key[];		/* (E) Koordinaten des Elements */
short   *ir;		/* Rueckmeldung: 0 = ok, 1 = Element vorh., -1 = Fehler */
{
   int h;
   short int k;
   TMelem *q;
   TMelpk *p;

   q = TMI_ele(t, key, &k, &h);

   if (k == LEER)
   {
      *ir = -1;

      if (t->typ == KDIM)
      {
	 if ((q = malloc(sizeof(TMelem))) == NULL )
	    return ( q );	      /* no core ? */

	 q->m = 0;
      }
      else
      {
	 if ((q = p = malloc(sizeof(TMelpk))) == NULL )
	    return ( p );	      /* no core ? */

	 p->nv = NULL;
      }

      *ir = 0;
      q->next = t->tm_elist->tm_elem[h];  t->tm_elist->tm_elem[h] = q;

      for ( k = 0; k < t->typ; k++ )
	 q->k[k] = key[k];
   }
   else
      *ir = 1;	      /* Element vorhanden */

   return ( q );
}

void
TMS_del (t)		      /* Loeschen einer Tabelle */
TMtab *t;		      /* (E) Verweis auf zu loeschende Tabelle */
{
   int i;
   int size;
   TMelem *p, *q;
   TMelpk *pp;

   size = t->tsize;

   for (i = 0;  i < size;  i++)
      for (p = t->tm_elist->tm_elem[i]; p; p = q )
      {
	 q = p->next;

         if (t->typ > KDIM)
         {
            pp = p;

            if (pp->nv)
               free(pp->nv);
         }

	 free(p);
      }

   free(t);
}

TMS_spk( k1, k2, k12 )	      /* Erzeugen eines Typ-6 Schluessels */
TMkey k1[], k2[], k12[];
{
   short i;
   int k;

   for ( i = 0; i < KDIM;  i++ )
      if ( k = k1[i] - k2[i] )
	 break;

   if ( k <= 0 )
      for ( i = 0; i < KDIM; i++ )
      {
	 k12[i] = k1[i];   k12[i+KDIM] = k2[i];
      }
   else
      for ( i = 0; i < KDIM; i++ )
      {
	 k12[i] = k2[i];   k12[i+KDIM] = k1[i];
      }
}

TMelem *		      /* Verweis auf Element, falls vorhanden */
TMA_ele (t, key, k)	      /* Element in Tabelle vorhanden ? */
TMtab *t;		      /* (E) Verweis auf aktuelle Tabelle */
TMkey key[];		      /* (E) Koordinaten des Elements */
short int *k;		      /* (A) Rueckmeldung: 0 = nicht vorhanden, 1 = vorhanden, -1 = Fehler */
{
   int h;
   TMelem *p;

   p = TMI_ele(t, key, k, &h);

   if (*k == LEER)
      p = t->tm_elist->tm_elem[h];

   t->wref++;
   return ( p );
}

TMelem *		   /* Verweis auf Element */
TMA_nxt (t, n, nn)	   /* Alle Elemente einer Tabelle durchmustern */
TMtab *t;		   /* (E) Verweis auf aktuelle Tabelle */
int   n,		   /* (E) interner Tabellenindex */
     *nn;		   /* (A) Rueckmeldung: < 0 = Tabelle erschoepft, >= 0 = interner Tabellenindex */
{
   int i, 
       j;
   static int size;
   static TMelem *p, *q;

   if (n < 0)
   {
      n = -1;
      size = t->tsize;
      q = NULL;
   }

   if (p = q)
   {
      q = p->next;
      *nn = n;
      return ( p );
   }

   for (i = n+1;  i < size;  i++)
      if (p = t->tm_elist->tm_elem[i])
      {
	 q = p->next;
	 *nn = i;
	 return ( p );
      }

   *nn = -1;
   return ( NULL );
}

void				 /* Statistische Werte zu einer Tabelle ermitteln */
TMA_stat(t, ne, wref, mxlk, nem, bf, mxw, mnw )
TMtab *t;			 /* (E) Verweis auf aktuelle Tabelle */
int   *ne,			 /* (A) Anzahl definierter Elemente */
      *wref,			 /* (A) Anzahl Zugriffe auf Tabellenwerte */
      *mxlk,			 /* (A) maximale Laenge der Elementketten */
      *nem;			 /* (A) Anzahl mehrfach belegter Eintraege */
float *bf,			 /* (A) Belegungsfaktor */
      *mxw, *mnw;		 /* (A) maximaler und minimaler Elementwert */
{
   int i, j, k, n, m,
       nte,		/* Anzahl von Tabellenelementen */
       nkk;		/* Anzahl von Kettenkoepfen */
   TMelem *p;

   *mxw = *mnw = .0;
   nte = nkk = 0;

   for (n = m = k = 0, i = -1;  (p = TMA_nxt(t, i, &j)) && j >= 0;  i = j)
   {
      if (t->typ <= 3)
	 if (p->w > *mxw)
	    *mxw = p->w;
	 else if (p->w < *mnw)
	    *mnw = p->w;

      if (i == j)	      /* Eintrag in Kette */
	 k++;
      else		      /* Eintrag in Tabelle */
      {
	 if (k)
	 {
	    if (k > m)
	       m = k;

	    nkk++;
	 }

	 k = 0;
	 nte++;
      }

      n++;
   }   

   if (k)
   {
      if (k > m)
	 m = k;

      nkk++;
   }

   *ne = n;
   *wref = t->wref;
   *mxlk = m+1;
   *nem  = nkk;
   *bf = (float) nte / (float) t->tsize;
}

TMelem *		      /* Verweis auf Element, falls vorhanden */
TMI_ele (t, key, k, ind)      /* Element in Tabelle vorhanden ? */
TMtab *t;		      /* (E) Verweis auf aktuelle Tabelle */
TMkey key[];		      /* (E) Koordinaten des Elements */
short int *k;		      /* (A) Rueckmeldung: 0 = nicht vorhanden, 1 = vorhanden */
int *ind;		      /* (A) Interner Tabellenindex, falls nicht vorhanden */
{
   short i, Lk;
   int ii, jj, h;
   TMtab *htab;
   TMelem *e;
   TMelpk *q;
   TMkey m; 
   unsigned short int dim;

   dim = t->tdim;

   for ( h = key[0], i = 1; i < dim; i++ )
   {
      m  = t->tmax[i];
      h = h * m + key[i];
   }

   h = h % t->tsize;

   for ( ii = 0; ii < t->tsize; ii++ )
   {
      for (Lk = 0, q = t->tm_elist->tm_elem[h];  q;  q = q->next, Lk++)
      {
	 for ( i = 0; i < t->typ && (q->k[i] == key[i]); i++ )
	    ;

	 if (i == t->typ)
	 {
	    *k = 1;   *ind = h;
	    return ( q );
	 }
      }

      if (Lk <= t->maxLk)
      {
	 *k = 0;   *ind = h;
	 return ( NULL );
      }

      h = (h + 1 < t->tsize ? h + 1 : 0);	/* Re-hash */
   }

   printf("\n++ TMI_ele: Tabelle voll. Reorganisation der TM-Tabelle (%d)!\n", t);

   for ( i = 0; i < dim; i++ )
      t->tmax[i] += 3;

   if (htab = TMS_def(t->typ, t->tmax))
   {
      TMelist   *tmp;

      for ( ii = -1; (q = TMA_nxt(t, ii, &jj)) && jj >= 0; ii = jj )
	 if ((e = TMI_ele(htab, q->k, &i, &h)) == NULL && i == 0)
	 {
	    t->tm_elist->tm_elem[jj] = q->next;     /* Element in alter Tabelle ausketten */
	    q->next = htab->tm_elist->tm_elem[h];   /* ... und in neue Tabelle einfuegen */
	    htab->tm_elist->tm_elem[h] = q;
	 }
	 else
	 {
	    printf("!! Speichermangel bei Tabellenreorganisation!\n");
	    exit();
	 }

      tmp = t->tm_elist;
      t->tm_elist = htab->tm_elist;
      t->tsize = htab->tsize;
      free(tmp);
      free(htab);
   }
   else
   {
      printf("!! Speichermangel bei Tabellenreorganisation!\n");
      exit();
   }

   return ( TMI_ele(t, key, k, ind));
}
